svg.railroad rect.railroad_canvas {
fill: hsl(30, 20%, 95%);
stroke-width: 0px;
}

svg.railroad path {
stroke-width: 3px;
stroke: black;
fill: none;
}

svg.railroad .debug {
stroke-width: 1px;
stroke: red;
}

svg.railroad text {
font: 14px monospace;
text-anchor: middle;
}

svg.railroad .nonterminal text {
font-weight: bold;
}

svg.railroad text.comment {
font: italic 12px monospace;
}

svg.railroad rect {
stroke-width: 3px;
stroke: black;
fill:hsl(-290, 70%, 90%);
}

svg.railroad g.labeledbox > rect {
stroke-width: 1px;
stroke: grey;
stroke-dasharray: 5px;
fill:rgba(90, 90, 150, .1);
}
